'''
--------------------------------------------------------------------------------

    toolbox.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

'''
The module toolbox provides general functions and constants needed by Lea classes 
'''

from math import log
from functools import wraps
import sys
import csv

def calcGCD(a,b):
    ''' returns the greatest common divisor between the given integer arguments
    '''
    while a > 0:
        (a,b) = (b%a,a)
    return b
    
def calcLCM(values):
    ''' returns the greatest least common multiple among the given sequence of integers
        requires that all values are strictly positive (not tested) 
    ''' 
    values1 = list(values)
    while len(set(values1)) > 1:
        minVal = min(values1)
        idx = values1.index(minVal)
        values1[idx] += values[idx]
    return values1[0]
    
LOG2 = log(2.)

def log2(x):
    ''' returns a float number that is the logarithm in base 2 of the given float x
    '''
    return log(x)/LOG2

def makeTuple(v):
    ''' returns a tuple with v as unique element
    '''    
    return (v,)
    
def easyMin(*args):
    ''' returns the minimum element of given args
        Note: if only one arg, it is returned (unlike Python's min function)
    '''
    if len(args) == 1:
        return args[0]
    return min(args)

def easyMax(*args):
    ''' returns the maximum element of given args
        Note: if only one arg, it is returned (unlike Python's max function)
    '''
    if len(args) == 1:
        return args[0]
    return max(args)

def genPairs(seq):
    ''' generates as tuples all the pairs from the elements of given sequence seq
    '''
    tuple1 = tuple(seq)
    length = len(tuple1)
    if length < 2:
        return
    if length == 2:
        yield tuple1
    else:
        head = tuple1[0]
        tail = tuple1[1:]
        for a in tail:
            yield (head,a)
        for pair in genPairs(tail):
            yield pair
                
# Python 2 / 3 dependencies
# the following redefines / rebinds the following objects in Python2: 
#  input
#  zip
#  next
#  dict
#  defaultdict
# these shall be imported by all modules that uses such names

# standard input function, zip and dictionary methods as iterators
from collections import defaultdict
if sys.version_info.major == 2:
    # Python 2.x
    # the goal of this part is to mimic a Python3 env in a Python2 env
    # rename raw_input method
    input = raw_input
    # zip as iterator shall be imported
    from itertools import izip as zip
    # next method shall be accessible as function
    def next(it):
        return it.next()
    # the dictionary classes shall have keys, values, items methods
    # wich are iterators; note that dictionaries must be created
    # with dict() instead of {}
    class dict(dict):
        keys = dict.iterkeys
        values = dict.itervalues
        items = dict.iteritems
    class defaultdict(defaultdict):
        keys = defaultdict.iterkeys
        values = defaultdict.itervalues
        items = defaultdict.iteritems
else:
    # Python 3.x
    # the following trick is needed to be able to import the names
    input = input
    zip = zip
    next = next
    dict = dict

def memoize(f):
   ''' returns a memoized version of the given instance method f;
       requires that the instance has a _cachesByFunc attribute
       referring to a dictionary;
       can be used as a decorator
       note: not usable on functions and static methods
   '''
   @wraps(f)
   def wrapper(obj,*args):
       # retrieve the cache for method f
       cache = obj._cachesByFunc.get(f)
       if cache is None:
           # first call to obj.f(...) -> build a new cache for f
           cache = obj._cachesByFunc[f] = dict()
       elif args in cache:
           # obj.f(*args) already called in the past -> returns the cached result
           return cache[args]
       # first call to obj.f(*args) -> calls obj.f(*args) and store the result in the cache    
       res = cache[args] = f(obj,*args)
       return res
   return wrapper

def strToBool(bStr):
    ''' returns True  if bStr is '1', 't', 'true', 'y' or 'yes' (case insentive)
                False if bStr is '0', 'f', 'false', 'n' or 'no' (case insentive)
        raise ValueError exception in other cases
    '''
    bStr = bStr.lower()
    if bStr in ('t','true','1','y','yes'):
        return True
    if bStr in ('f','false','0','n','no'):
        return False
    raise ValueError("invalid boolean litteral '%s'"%bStr)

def readCSVFilename(csvFilename,colNames=None,dialect='excel',**fmtparams):
    ''' same as readCSVFile method, except that it takes a filename instead
        of an open file (i.e. the method opens itself the file for reading);
        see readCSVFile doc for more details
    '''
    with open(csvFilename,'rU') as csvFile:
        return readCSVFile(csvFile,colNames,dialect,**fmtparams)

def readCSVFile(csvFile,colNames=None,dialect='excel',**fmtparams):
    ''' returns a tuple (attrNames,dataFreq) from the data read in the given CSV file
        * attrNames is a tuplewith the attribute names found in the header row 
        * dataFreq is a list of tuples (tupleValue,count) for each CSV row 
          with tupleValue containing read fields and count the positive integer
          giving the probability weight of this row;
        the arguments follow the same semantics as those of Python's csv.reader
        method, which supports different CSV formats
        see doc in https://docs.python.org/2/library/csv.html
        * if colNames is None, then the fields found in the first read row of the CSV
          file provide information on the attributes: each field is made up of a name,
          which shall be a valid identifier, followed by an optional 3-characters type
          code among  
            {b} -> boolean
            {i} -> integer
            {f} -> float
            {s} -> string
            {#} -> count   
          if the type code is missing for a given field, the type string is assumed for
          this field; for example, using the comma delimiter (default), the first row
          in the CSV file could be:
              name,age{i},heigth{f},married{b}
        * if colNames is not None, then colNames shall be a sequence of strings giving
          attribute information as described above, e.g.
              ('name','age{i}','heigth{f}','married{b}')
          it assumed that there is NO header row in the CSV file
        the type code defines the conversion to be applied to the fields read on the
        data lines; if the read value is empty, then it is converted to Python's None,
        except if the type is string, then, the value is the empty string; 
        if the read value is not empty and cannot be parsed for the expected type, then
        an exception is raised; for boolean type, the following values (case
        insensitive):
          '1', 't', 'true', 'y', 'yes' are interpreted as Python's True,
          '0', 'f', 'false', 'n', 'no' are interpreted as Python's False;
        the {#} code identifies a field that provides a count number of the row,
        representing the probability of the row or its frequency as a positive integer;
        such field is NOT included as attribute of the joint distribution; it is useful
        to define non-uniform probability distribution, as alternative to repeating the
        same row multiple times
    '''
    # read the CSV file
    attrNames = []
    convFunctions = []
    countAttrIdx = None
    fieldsPerRowIter = csv.reader(csvFile,dialect,**fmtparams)
    if colNames is None:
        # parse the header row
        colNames = next(fieldsPerRowIter)
    # if colNames is not None, it is assumed that there is no header row in the CSV file
    for (colIdx,colName) in enumerate(colNames):
        colName = colName.strip()
        if colName.endswith('{#}'):
            if countAttrIdx is not None:
                raise ValueError("count column ('{#}') must be unique in CSV header line")
            countAttrIdx = colIdx
        else:
            hasSuffix = True
            convFunction = None    
            if colName.endswith('{b}'):
                convFunction = strToBool
            elif colName.endswith('{i}'):
                convFunction = int
            elif colName.endswith('{f}'):
                convFunction = float
            elif not colName.endswith('{s}'):
                hasSuffix = False
            if hasSuffix:
                attrName = colName[:-3].strip()
            else:
                attrName = colName
            attrNames.append(attrName)
            convFunctions.append(convFunction)
    # parse the data rows
    fieldsPerRow = tuple(fieldsPerRowIter)
    dataFreq = []
    for fields in fieldsPerRow:
        if countAttrIdx is None:
            # no 'count' field: each read row has a count of 1 
            count = 1
        else:
            # 'count' field: extract the count value from the fields
            count = int(fields.pop(countAttrIdx))
        # conversion of read fields according to optional given types
        convFields = []
        for (field,convFunction) in zip(fields,convFunctions):
            if convFunction is None:
                convField = field
            else:
                if field == '':
                    # empty value translated as Python's None 
                    convField = None
                else:
                    convField = convFunction(field)
            convFields.append(convField)
        dataFreq.append((tuple(convFields),count))
    return (attrNames,dataFreq)




'''
--------------------------------------------------------------------------------

    leapp_console.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

# leapp console configuration parameters
PROMPT1 = 'lea> '
PROMPT2 = ' ... '
PROMPTD = ' py# '
DEBUG = False

CONTINUATION_LINE_CHARS = '(,:\\'

from leapp_translator import LeappTranslator
import license
from toolbox import input

import traceback

try:
    # try to activate terminal history 
    import readline
except:
    # OS dependent failure -> no terminal history
    pass
            
class LeappConsole(object):
    
    ''' 
    A LeappConsole instance allows to create an interactive console where the user
    can type leapp statements, which are translated into Python statements using
    Lea module and executed on the-fly.
    '''
        
    __slots__ = ( 'locals', 'debug', 'prompt1', 'prompt2', 'promptD')
    
    def __init__(self):
        self.locals = locals()
        self.debug = DEBUG
        self.prompt1 = PROMPT1
        self.prompt2 = PROMPT2
        self.promptD = PROMPTD
        self.execPythonStatement('_leapp = self')
        self.execPythonStatement('del self')
        self.execPythonStatement('from lea import *')

    class Error(Exception):
        pass

    def execLeappTranslatorMultilineStatement(self,rMultilineStatement):
        pMultilineStatement = LeappTranslator.getTarget00(rMultilineStatement)
        if self.debug:
            print (self.promptD+('\n'+self.promptD).join(pMultilineStatement.split('\n')))
        self.execPythonStatement(pMultilineStatement)

    def execPythonStatement(self,pStatement):
        if len(pStatement) > 0:
            code = compile(pStatement+'\n','<leapp>','single')
            exec (code,self.locals)

    def inputMultilineStatement(self):
        readLines = []
        prompt = self.prompt1
        while True:
            readLine = input(prompt).rstrip()
            if len(readLine) == 0:
                break
            lastChar = readLine[-1]
            if lastChar == ':' and len(readLine) > 1:
                readLine += '\n'
            elif lastChar == '\\':
                readLine = readLine[:-1]
            readLines.append(readLine)
            if len(readLine) <= 1 or lastChar not in CONTINUATION_LINE_CHARS:
                break
            prompt = self.prompt2
        return ''.join(readLines)    

    def startCmdLoop(self):
        from lea import Lea
        while True:
            try:
                rMultilineStatement = self.inputMultilineStatement()
            except EOFError:
                break
            try:
                try:
                    self.execLeappTranslatorMultilineStatement(rMultilineStatement)
                except:
                    if self.debug:
                        traceback.print_exc()
                    raise  
            except Lea.Error as exc:
                print ("Lea error: %s"%exc)
            except LeappTranslator.Error as exc:
                print ("Leapp syntax error: %s"%exc)
            except Exception as exc:
                print ("Python error: %s"%exc)




'''
--------------------------------------------------------------------------------

    flea.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea
from clea import Clea

class Flea(Lea):
    
    '''
    Flea is a Lea subclass, which instance is defined by a function applied on a given sequence
    of arguments. The arguments are coerced to Lea instances. The function is applied on all elements
    of cartesian product of all arguments (see Clea class). This results in a new probability
    distribution for all the values returned by the function.
    '''
    
    __slots__ = ('_f','_cleaArgs')

    def __init__(self,f,cleaArgs):
        Lea.__init__(self)
        self._f = f
        self._cleaArgs = cleaArgs
    
    @staticmethod
    def build(f,args):
        return Flea(f,Clea(*args))

    def _getLeaChildren(self):
        return (self._cleaArgs,)

    def _clone(self,cloneTable):
        return Flea(self._f,self._cleaArgs.clone(cloneTable))    

    def _genVPs(self):
        f = self._f
        if isinstance(f,Lea):
            for ((f2,args),p) in Clea(f,self._cleaArgs)._genVPs():
                yield (f2(*args),p)            
        else:
            for (args,p) in self._cleaArgs._genVPs():
                yield (f(*args),p)

    def _genOneRandomMC(self):
        f = self._f
        if isinstance(f,Lea):
            for (f2,args) in Clea(f,self._cleaArgs)._genOneRandomMC():
                yield f2(*args)            
        else:
            for args in self._cleaArgs._genOneRandomMC():
                yield f(*args)



'''
--------------------------------------------------------------------------------

    plea.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea
from toolbox import makeTuple

class Plea(Lea):

    '''
    Plea is a Lea subclass, which instance represents a probability distribution obtained by making 
    a cartesian product of given Lea instancea with itself, a given number of times. Each value is 
    a tuple having the given number as size.
    '''

    __slots__ = ('_lea1','_lea1Tuple','_nTimes')

    def __init__(self,lea1,nTimes=2):
        Lea.__init__(self)
        self._lea1 = lea1
        self._lea1Tuple = lea1.map(makeTuple)
        self._nTimes = nTimes
        if nTimes <= 0:
            raise Lea.Error("cprodTimes method requires a strictly positive integer")

    def _getLeaChildren(self):
        return (self._lea1,)

    def _clone(self,cloneTable):
        return Plea(self._lea1.clone(cloneTable),self._nTimes)
    
    def _genVPs(self,nTimes=None):
        if nTimes is None:
            nTimes = self._nTimes
        if nTimes == 1:
            return self._lea1Tuple._genVPs()
        # nTimes >= 2 : use dichotomic algorithm
        nTimes1 = nTimes // 2
        plea = Plea(self._lea1,nTimes1)
        alea = plea.getAlea()
        flea = alea + alea.clone()
        if nTimes%2 == 1:
            # nTimes is odd : nTimes = 2*nTimes1 + 1
            # operate with one more lea1 on the current result 
            flea += self._lea1Tuple
        return flea._genVPs()

    def _genOneRandomMC(self):
        yield tuple(v for _ in range(self._nTimes) for v in self._lea1._genOneRandomMC())



'''
--------------------------------------------------------------------------------

    ilea.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea
from operator import and_

class Ilea(Lea):
    
    '''
    Ilea is a Lea subclass, which instance represents a probability distribution obtained
    by filtering the values Vi of a given Lea instance that verify a boolean condition C,
    which is the AND of given boolean conditions Cj(Vi).
    In the context of a conditional probability table (CPT), each Ilea instance represents
    a given distribution <Vi,p(Vi|C)>, assuming that a given condition C is verified (see Blea class).
    '''

    __slots__ = ('_lea1','_condLeas')

    def __init__(self,lea1,condLeas):
        Lea.__init__(self)
        self._lea1 = lea1
        self._condLeas = tuple(condLeas)

    def _getLeaChildren(self):
        return (self._lea1,) + self._condLeas
    
    def _clone(self,cloneTable):
        return Ilea(self._lea1.clone(cloneTable),(condLea.clone(cloneTable) for condLea in self._condLeas))

    @staticmethod
    def _genTrueP(condLeas):
        ''' generates probabilities of True for ANDing the given conditions 
            this uses short-circuit evaluation
        '''
        if len(condLeas) == 0:
            # empty condition: evaluated as True (seed of recursion)
            yield 1
        else:
            for (cv0,p0) in condLeas[0]._genVPs():
                if cv0 is True:
                    # the first condition is true, for some binding of variables
                    for p1 in Ilea._genTrueP(condLeas[1:]):
                        # the full condition is true, for some binding of variables
                        yield p0*p1
                elif cv0 is False:
                    # short-circuit: do not go further since the AND is false
                    pass
                else:
                    # neither True, nor False -> error
                    raise Lea.Error("boolean expression expected")
    
    def _genVPs(self):
        for cp in Ilea._genTrueP(self._condLeas):
            # the AND of conditions is true, for some binding of variables
            # yield value-probability pairs of _lea1, given this binding
            for (v,p) in self._lea1._genVPs():
                yield (v,cp*p)

    def _genOneRandomTrueCond(self,condLeas,withException):
        if len(condLeas) == 0:
            # empty condition: evaluated as True (seed of recursion)
            yield None
        else:
            for cv in condLeas[0]._genOneRandomMC():
                if cv is True:
                    for v in self._genOneRandomTrueCond(condLeas[1:],withException):
                        yield v
                elif cv is False:
                    if withException:
                        raise Lea._FailedRandomMC()
                    yield self
                else:
                    raise Lea.Error("boolean expression expected")

    def _genOneRandomMC(self):
        for _ in self._genOneRandomTrueCond(self._condLeas,True):
            for v in self._lea1._genOneRandomMC():
                yield v

    def _genOneRandomMCNoExc(self):
        for u in self._genOneRandomTrueCond(self._condLeas,False):
            if u is not self: 
                for v in self._lea1._genOneRandomMC():
                    yield v

    def lr(self):
        ''' returns a float giving the likelihood ratio (LR) of an 'evidence' E,
            which is self's unconditional probability distribution, for a given
            'hypothesis' H, which is self's condition; it is calculated as 
                  P(E | H) / P(E | not H)
            both E and H must be boolean probability distributions, otherwise
            an exception is raised;
            an exception is raised also if H is certainly true or certainly false      
        '''
        lrN = self.P
        lrD = self._lea1.given(~Lea.reduce(and_,self._condLeas,False)).P
        if lrD == 0:
            if lrN == 0:
                raise Lea.Error("undefined likelihood ratio")
            return float('inf') 
        return float(lrN) / float(lrD)


'''
--------------------------------------------------------------------------------

    flea1.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea

class Flea1(Lea):
    
    '''
    Flea1 is a Lea subclass, which instance is defined by a function applied on one given Lea argument.
    The function is applied on all values of the argument. This results in a new probability distribution
    for all the values returned by the function.
    '''
    
    __slots__ = ('_f','_leaArg')

    def __init__(self,f,leaArg):
        Lea.__init__(self)
        self._f = f
        self._leaArg = leaArg

    def _getLeaChildren(self):
        return (self._leaArg,)

    def _clone(self,cloneTable):
        return Flea1(self._f,self._leaArg.clone(cloneTable))    

    def _genVPs(self):
        f = self._f
        for (v,p) in self._leaArg._genVPs():
            yield (f(v),p)

    def _genOneRandomMC(self):
        for v in self._leaArg._genOneRandomMC():
            yield self._f(v)



'''
--------------------------------------------------------------------------------

    prob_fraction.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from toolbox import calcLCM

from fractions import Fraction

class ProbFraction(Fraction):

    ''' 
    A ProbFraction instance represents a probability as a fraction
    It inherits Fraction, overloading methods to improve useability
    '''
    
    class Error(Exception):
        pass

    def __new__(cls, numerator=0, denominator=None):
        ''' returns a new instance of ProbFraction
            following signatures of Fraction constructor
            + allowing a percentage in numerator as a string 'xxx %'
              with xxx being a float literal
            Note that the constructor does NOT check that the fraction
            is in the range 0 to 1; this is so to allow intermediate
            results in expressions to go beyond that range;
            the range is verified when string representation is required
            (method str) or by explicit call to check() method 
        '''
        if isinstance(numerator,str) and denominator is None:
            numerator = numerator.strip()
            if numerator.endswith('%'):
                numerator = Fraction(numerator[:-1])
                denominator = 100
        fraction = Fraction(numerator,denominator)        
        return ProbFraction.fromFraction(fraction)
    
    @staticmethod         
    def fromFraction(fraction):
        ''' static method, returns a ProbFraction numerically
            equivalent to the given Fraction instance
        '''
        probFraction = Fraction.__new__(ProbFraction)
        probFraction._numerator = fraction._numerator
        probFraction._denominator = fraction._denominator
        return probFraction

    @staticmethod
    def coerce(value):
        ''' static method, returns a ProbFraction instance corresponding the given value:
            if the value is a ProbFraction instance, then it is returned
            otherwise, a new ProbFraction instance is returned corresponding to given value
        '''
        if not isinstance(value,ProbFraction):
            value = ProbFraction(value)
        return value
        
    def __coerceFunc(f):
        ''' internal utility function
            returns a function returning a ProbFraction
            equivalent to the given function f returning Fraction
        '''
        return lambda *x: ProbFraction.fromFraction(f(*x))
     
    # overloading arithmetic magic methods of Fraction
    # to convert Fraction result into ProbFraction result
    # Note: do not overwrite __floordiv__, __rfloordiv__, __pow__
    # since these methods do not return Fraction instances
    __pos__      = __coerceFunc(Fraction.__pos__)
    __neg__      = __coerceFunc(Fraction.__neg__)
    __pow__      = __coerceFunc(Fraction.__pow__)
    __add__      = __coerceFunc(Fraction.__add__)
    __radd__     = __coerceFunc(Fraction.__radd__)
    __sub__      = __coerceFunc(Fraction.__sub__)
    __rsub__     = __coerceFunc(Fraction.__rsub__)
    __mul__      = __coerceFunc(Fraction.__mul__)
    __rmul__     = __coerceFunc(Fraction.__rmul__)
    __truediv__  = __coerceFunc(Fraction.__truediv__)
    __rtruediv__ = __coerceFunc(Fraction.__rtruediv__)

    # Python 2 compatibility
    __div__ = __truediv__
    __rdiv__ = __rtruediv__

    def check(self):
        ''' raises an Exception if the fraction is not in the probability range, from 0 to 1
        '''
        if not (0 <= self <= 1):
            raise ProbFraction.Error("%s is not a valid probability (between 0 and 1)"%Fraction.__str__(self))

    def __str__(self):
        ''' returns a string representation "numerator/denominator" of self
            raises an Exception if the fraction is not in the probability range, from 0 to 1
        '''
        self.check()
        return Fraction.__str__(self)
    
    # overwrites representation method
    __repr__ = __str__
    
    def asFloat(self):
        ''' returns float string representation of self
            note: to get float number, use float(self) 
        '''
        return str(float(self))

    def asPct(self):
        ''' returns float percentage string representation of self
        '''
        return "%f %%" % float(self*100)
    
    @staticmethod
    def getProbWeights(fractions):
        ''' static method, returns a tuple of integers
            which are the numerators of given sequence of fractions,
            after conversion to a common denominator  
        '''
        denominators = tuple(fraction.denominator for fraction in fractions)
        if len(denominators) == 0:
            raise ProbFraction.Error('getProbWeights requires at least one fraction')
        lcm = calcLCM(denominators)
        return tuple(fraction.numerator*(lcm//fraction.denominator) for fraction in fractions) 
